package net.peachjean.guice.command.invoke;

import com.google.inject.*;
import com.google.inject.name.Names;
import net.peachjean.guice.command.ArgIterator;
import net.peachjean.guice.command.CommandScoped;
import net.jcip.annotations.NotThreadSafe;

import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Type;
import java.util.Iterator;
import java.util.NoSuchElementException;

@NotThreadSafe
@CommandScoped
class ArgIteratorImpl implements ArgIterator {

	static final String NUM_ARGS = "__NUM_ARGS__";
	private int numArgs = 0;
	private int current = 0;
	private Injector argInjector;

	@Inject
	ArgIteratorImpl(Injector injector, final CommandParser commandParser) {
		this.argInjector = injector.createChildInjector(new AbstractModule() {

			@Override
			protected void configure() {
				Iterator<String> argIterator = commandParser.provideArgIterator();
				int index = 0;
				while (argIterator.hasNext()) {
					ArgIteratorImpl.bindArg(binder(), argIterator.next(), index++);
				}
				ArgIteratorImpl.bindNumArgs(binder(), index);
			}
		});
		this.numArgs = this.argInjector.getInstance(Key.get(int.class, Names.named(NUM_ARGS)));
	}

	@Override
	public boolean hasMoreArgs() {
		return current < numArgs;
	}

	@Override
	public Object next(Type type) throws NoSuchElementException {
		if (current < numArgs) {
			return argInjector.getInstance(Key.get(type, new IndexedArgImpl(current++)));
		} else {
			throw new NoSuchElementException("No more args are available.");
		}
	}

	static void bindArg(Binder binder, String value, int index) {
		binder.bindConstant().annotatedWith(new IndexedArgImpl(index)).to(value);
	}

	static void bindNumArgs(Binder binder, int value) {
		binder.bind(Integer.class).annotatedWith(Names.named(NUM_ARGS)).toInstance(value);
	}

	@SuppressWarnings({"ClassExplicitlyAnnotation"})
	static class IndexedArgImpl implements IndexedArg {
		private int value;

		public IndexedArgImpl(int value) {
			this.value = value;
		}

		@Override
		public int value() {
			return value;
		}

		@Override
		public Class<? extends Annotation> annotationType() {
			return IndexedArg.class;
		}

		@Override
		public boolean equals(Object o) {
			return o instanceof IndexedArg && ((IndexedArg) o).value() == value();
		}

		@Override
		public int hashCode() {
			return (127 * "value".hashCode()) ^ Integer.valueOf(value).hashCode();
		}

		@Override
		public String toString() {
			return String.format("<Argument %d>", value);
		}
	}

	@BindingAnnotation
	@Retention(RetentionPolicy.RUNTIME)
	@Target({})
	static @interface IndexedArg {
		int value();
	}
}
