package net.peachjean.guice.command.invoke;

import com.google.inject.*;
import net.peachjean.guice.command.Arg;
import net.peachjean.guice.command.CommandConfigurationException;
import net.peachjean.guice.command.CommandInvocationException;
import net.peachjean.guice.command.Out;
import net.jcip.annotations.NotThreadSafe;

import java.io.PrintStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Type;

@NotThreadSafe
class CommandInvokerImpl implements CommandInvoker {

	static final String ERROR_PRIVATE_COMMAND = "Command %s could not be executed.  Command methods must be public.";
	static final String ERROR_COMMAND_GENERIC = "Command %s could not be executed.";
	private Injector injector;
	private Provider<PrintStream> out;
	private ArgResolverRegistry argResolverRegistry;
	private Provider<CommandParser> parserProvider;
	private CommandScope commandScope;

	@Inject
	CommandInvokerImpl(Injector injector, @Out Provider<PrintStream> out, ArgResolverRegistry argResolverRegistry,
	                   Provider<CommandParser> parserProvider, CommandScope commandScope) {
		this.injector = injector;
		this.out = out;
		this.argResolverRegistry = argResolverRegistry;
		this.parserProvider = parserProvider;
		this.commandScope = commandScope;
	}

	@Override
	public void invoke(CommandInvocation commandInvocation) {
		GuiceAppContext guiceAppContext = new GuiceAppContext(commandInvocation);
		commandScope.performInScope(guiceAppContext, new Runnable() {
			@Override
			public void run() {
				final CommandDescriptor descriptor = determineCommand();
				Object results = buildAndInvokeCommand(descriptor);
				handleCommandResults(results, descriptor);
			}
		});
	}

	private CommandDescriptor determineCommand() {
		return parserProvider.get().provideCommandDescriptor();
	}

	private Object buildAndInvokeCommand(CommandDescriptor descriptor) {
		Object commandObject = injector.getInstance(descriptor.key);
		Object[] commandParameters = buildParameterArray(descriptor, injector);
		return invokeCommand(descriptor, commandObject, commandParameters);
	}

	private Object invokeCommand(CommandDescriptor descriptor, Object commandObject, Object[] commandParameters) {
		try {
			return descriptor.method.invoke(commandObject, commandParameters);
		} catch (IllegalAccessException e) {
			// this should never be reached - a private method should not be registered as a command in the first place.
			throw new CommandConfigurationException(ERROR_PRIVATE_COMMAND, e, descriptor.command);
		} catch (InvocationTargetException e) {
			throw new CommandInvocationException(String.format(ERROR_COMMAND_GENERIC, descriptor.command), e);
		}
	}

	private void handleCommandResults(Object results, CommandDescriptor commandDescriptor) {
		if (!Void.TYPE.isAssignableFrom(commandDescriptor.method.getReturnType())) {
			this.out.get().println(results);
		}
	}

	private Object[] buildParameterArray(CommandDescriptor descriptor, Injector injector) {
		final Method method = descriptor.method;
		final Object[] parameterArray = new Object[method.getParameterTypes().length];

		for (int i = 0; i < method.getParameterTypes().length; i++) {
			if (parameterHasAnnotation(method, i, Arg.class)) {
				parameterArray[i] = argResolverRegistry.resolveArgValue(method, i);
			} else {
				parameterArray[i] = resolveSimpleValue(injector, method, i);
			}
		}
		return parameterArray;
	}

	private Object resolveSimpleValue(Injector injector, Method method, int parameterIndex) {
		return injector.getInstance(getParameterKey(method, parameterIndex));
	}

	static Key<?> getParameterKey(Method method, int i) {
		Type type = method.getGenericParameterTypes()[i];
		for (Annotation annotation : method.getParameterAnnotations()[i]) {
			if (annotation.annotationType().getAnnotation(BindingAnnotation.class) != null) {
				return Key.get(type, annotation);
			}
		}
		return Key.get(type);
	}

	static boolean parameterHasAnnotation(Method method, int parameter, Class<? extends Annotation> annotation) {
		for (Annotation parameterAnnotation : method.getParameterAnnotations()[parameter]) {
			if (annotation.isInstance(parameterAnnotation)) {
				return true;
			}
		}
		return false;
	}

}
