package net.peachjean.guice.command.invoke;

import com.google.common.collect.Iterators;
import com.google.common.collect.PeekingIterator;
import com.google.inject.Inject;
import net.peachjean.guice.command.*;
import org.apache.commons.cli.*;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;

import java.util.Iterator;
import java.util.Map;

@CommandScoped
class CommandParser {

	private String[] args;
	private CommandRegistry commandRegistry;
	private CommandLineParser commandLineParser;
	private OptionInfoRegistry optionInfoRegistry;

	private boolean parsed = false;

	// parse results

	private CommandDescriptor commandDescriptor;
	private PeekingIterator<String> argIterator;
	private CommandLine commandLine;


	@Inject
	public CommandParser(GuiceAppContext guiceAppContext, CommandRegistry commandRegistry, CommandLineParser commandLineParser, OptionInfoRegistry optionInfoRegistry) {
		this.commandLineParser = commandLineParser;
		this.optionInfoRegistry = optionInfoRegistry;
		this.args = guiceAppContext.args;
		this.commandRegistry = commandRegistry;
	}

	public CommandDescriptor provideCommandDescriptor() {
		parseIfNeeded();
		return this.commandDescriptor;
	}

	public <T> T provideOptionValue(OptionKey<T> optionKey) {
		parseIfNeeded();
		OptionInfo<T> info = optionInfoRegistry.lookup(optionKey);
		return OptionType.determineType(info).getValue(info, commandLine);
	}

	private void parseIfNeeded() {
		if (!parsed) {
			parse();
			this.parsed = true;
		}
	}

	private void parse() {
		PeekingIterator<String> initialIterator = Iterators.peekingIterator(Iterators.forArray(args));
		this.commandDescriptor = determineCommand(commandRegistry.getDescriptorMap(), initialIterator);

		Options options = createOptions(this.commandDescriptor, this.commandRegistry);
		try {
			commandLine = commandLineParser.parse(options, Iterators.toArray(initialIterator, String.class));
			argIterator = Iterators.peekingIterator(commandLine.getArgList().iterator());
		} catch (ParseException e) {
			throw new MalformedCommandException("Could not parse options.", e);
		}
	}

	private Options createOptions(CommandDescriptor commandDescriptor, CommandRegistry commandRegistry) {
		Options options = new Options();
		for (OptionKey optionKey : commandRegistry.getApplicableOptions(commandDescriptor.command)) {
			options.addOption(createOption(optionKey));
		}
		return options;
	}

	private CommandDescriptor determineCommand(Map<String, CommandDescriptor> commandDescriptorMap, PeekingIterator<String> args) {
		final String command;
		if (!args.hasNext()) {
			if (commandDescriptorMap.containsKey(CommandDescriptor.DEFAULT_COMMAND)) {
				command = CommandDescriptor.DEFAULT_COMMAND;
			} else {
				throw new MalformedCommandException("A command must be provided.");
			}
		} else if (commandDescriptorMap.containsKey(args.peek())) {
			command = args.next();
		} else if (commandDescriptorMap.containsKey(CommandDescriptor.DEFAULT_COMMAND)) {
			command = CommandDescriptor.DEFAULT_COMMAND;
		} else {
			throw new MalformedCommandException(String.format("The command %s was not recognized and no default command is supported.", args.peek()));
		}
		return commandDescriptorMap.get(command);
	}

	public Iterator<String> provideArgIterator() {
		return argIterator;
	}

	private Option createOption(OptionKey optionKey) {
		OptionInfo info = optionInfoRegistry.lookup(optionKey);
		return OptionType.determineType(info).createOption(info);
	}

	static enum OptionType {
		BOOLEAN {
			@Override
			<T> T getValue(OptionInfo<T> info, CommandLine commandLine) {
				return (T) Boolean.valueOf(commandLine.hasOption(optionName(info)) ^ (Boolean) info.defaultValue);
			}

			@Override
			boolean matches(OptionInfo<?> info) {
				return Boolean.class.isAssignableFrom(info.type) || info.type == Boolean.TYPE;
			}

			protected void setupOption(OptionInfo<?> info, Option option) {
				option.setOptionalArg(false);
				option.setArgs(0);
			}
		},
		OTHER {
			@Override
			<T> T getValue(OptionInfo<T> info, CommandLine commandLine) {
				if (commandLine.hasOption(optionName(info))) {
					try {
						return (T) commandLine.getParsedOptionValue(optionName(info));
					} catch (ParseException e) {
						throw new MalformedCommandException("Could not parse command line.", e);
					}
				} else {
					return info.defaultValue;
				}
			}

			@Override
			boolean matches(OptionInfo<?> info) {
				return true;
			}

			protected void setupOption(OptionInfo<?> info, Option option) {
				option.setType(info.type);
				option.setOptionalArg(info.defaultValue != null);
			}
		};

		protected abstract void setupOption(OptionInfo<?> info, Option option);

		final Option createOption(OptionInfo<?> info) {
			Option option = createBaseOption(info);
			setupOption(info, option);
			return option;
		}

		abstract <T> T getValue(OptionInfo<T> info, CommandLine commandLine);

		abstract boolean matches(OptionInfo<?> info);

		private static Option createBaseOption(OptionInfo<?> info) {
			return new Option(info.shortValue, info.name, true, info.description);
		}

		private static <T> String optionName(OptionInfo<T> info) {
			return info.name;
		}

		public static OptionType determineType(OptionInfo<?> info) {
			for (OptionType type : values()) {
				if (type.matches(info)) {
					return type;
				}
			}
			throw new IllegalStateException("This point should never be reached.  Bug in " + OTHER);
		}
	}
}
