package net.peachjean.guice.command.invoke;

import com.google.common.collect.ImmutableMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import com.google.inject.Binding;
import com.google.inject.Key;
import com.google.inject.Module;
import com.google.inject.spi.DefaultElementVisitor;
import com.google.inject.spi.Element;
import com.google.inject.spi.Elements;
import net.peachjean.guice.command.*;
import net.peachjean.guice.command.app.GuiceApp;
import net.peachjean.guice.walker.DependencyVisitor;
import net.peachjean.guice.walker.DependencyWalker;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.*;

class CommandRegistryBuilder {

	static final CommandElementVisitor COMMAND_ELEMENT_VISITOR = new CommandElementVisitor();

	public CommandRegistryBuilder() {
	}

	CommandRegistry buildCommandRegistry(Iterable<? extends Module> modules) {
		final Map<String, CommandDescriptor> descriptorMap = buildCommandMap(modules);

		Multimap<String, OptionKey<?>> options = buildOptionMap(descriptorMap, modules);

		return new CommandRegistry(descriptorMap, options);
	}

	private Multimap<String, OptionKey<?>> buildOptionMap(Map<String, CommandDescriptor> descriptorMap, Iterable<? extends Module> modules) {
		final ImmutableMultimap.Builder<String, OptionKey<?>> builder = ImmutableMultimap.builder();
		for (final CommandDescriptor descriptor : descriptorMap.values()) {
			final Annotation[][] parameterAnnotations1 = descriptor.method.getParameterAnnotations();
			for (int i = 0; i < parameterAnnotations1.length; i++) {
				if (!CommandInvokerImpl.parameterHasAnnotation(descriptor.method, i, Arg.class)) {
					locateOptions(modules, builder, descriptor.command, CommandInvokerImpl.getParameterKey(descriptor.method, i));
				}
				for (Annotation annotation : parameterAnnotations1[i]) {
					if (annotation instanceof Option) {
						builder.put(descriptor.command, new OptionKey<Object>(((Option) annotation).value(),
						                                                      (Class<Object>) descriptor.method.getParameterTypes()[i]));
					}
				}
			}
			locateOptions(modules, builder, descriptor.command, descriptor.key);
		}
		return builder.build();
	}

	private void locateOptions(Iterable<? extends Module> modules, final ImmutableMultimap.Builder<String, OptionKey<?>> builder, final String command, Key<?> key) {
		DependencyWalker dependencyWalker = new DependencyWalker(key, modules);
		dependencyWalker.walk(new DependencyVisitor() {
			@Override
			public void visit(Key<?> key) {
				if (key.getAnnotation() instanceof Option) {
					builder.put(command, new OptionKey<Object>(((Option) key.getAnnotation()).value(),
					                                           (Class<Object>) key.getTypeLiteral().getRawType()));
				}
			}
		});
	}

	private Map<String, CommandDescriptor> buildCommandMap(Iterable<? extends Module> modules) {
		final Map<String, CommandDescriptor> commandDescriptorMap = new TreeMap<String, CommandDescriptor>();
		final List<Element> elements = Elements.getElements(modules);
		for (Element element : elements) {
			for (CommandDescriptor descriptor : element.acceptVisitor(COMMAND_ELEMENT_VISITOR)) {
				final CommandDescriptor otherDescriptor = commandDescriptorMap.put(descriptor.command, descriptor);
				if (otherDescriptor != null) {
					throw generateDuplicateCommandException(otherDescriptor, descriptor);
				}
			}
		}
		return commandDescriptorMap;
	}

	private CommandConfigurationException generateDuplicateCommandException(CommandDescriptor otherDescriptor, CommandDescriptor descriptor) {
		final String message;
		if (CommandDescriptor.DEFAULT_COMMAND.equals(descriptor.command)) {
			message = "Multiple default commands found. (%s.%s and %s.%s)";
		} else {
			message = "Multiple commands named " + descriptor.command + " were found. (%s.%s and %s.%s)";
		}
		return new CommandConfigurationException(message,
		                                         otherDescriptor.key.getTypeLiteral().getRawType().getSimpleName(),
		                                         otherDescriptor.method.getName(),
		                                         descriptor.key.getTypeLiteral().getRawType().getSimpleName(),
		                                         descriptor.method.getName());
	}

	private static class CommandElementVisitor extends DefaultElementVisitor<Iterable<CommandDescriptor>> {
		@Override
		public Iterable<CommandDescriptor> visit(Binding binding) {
			Class<?> candidateType = binding.getKey().getTypeLiteral().getRawType();
			Collection<CommandDescriptor> commands = new ArrayList<CommandDescriptor>();
			for (Method method : candidateType.getMethods()) {
				matchAndStoreCommand(binding, method, commands);
			}
			return commands;
		}

		private void matchAndStoreCommand(Binding binding, Method method, Collection<CommandDescriptor> commands) {
			if (method.isAnnotationPresent(Command.class)) {
				Command command = method.getAnnotation(Command.class);
				commands.add(new CommandDescriptor(command.value(), binding.getKey(), method));
			} else if (method.isAnnotationPresent(DefaultCommand.class)) {
				commands.add(new CommandDescriptor(CommandDescriptor.DEFAULT_COMMAND, binding.getKey(), method));
			}
		}
	}
}